#include "functions.h"

void print_vector(char s[], gsl_vector* v) {
	printf("%s\n",s);
	for(int i=0;i<v->size;i++) {
		printf("%15g",gsl_vector_get(v,i));
		printf("\n");
	}
}

void print_matrix(gsl_matrix* M, FILE* stream) {
    for(int i = 0; i < M->size1; i++) {
        for(int j = 0; j < M->size2; j++) {
            fprintf(stream,"%15g\t", gsl_matrix_get(M, i, j));
        }
        fprintf(stream,"\n");
    }
}

void timesJ(gsl_matrix* A, int p, int q, double theta) {
    double c=cos(theta),s=sin(theta);
    for(int i=0;i<A->size1;i++) {
        double Aip = gsl_matrix_get(A,i,p)*c-gsl_matrix_get(A,i,q)*s;
        double Aiq = gsl_matrix_get(A,i,p)*s+gsl_matrix_get(A,i,q)*c;
        gsl_matrix_set(A,i,p,Aip);
        gsl_matrix_set(A,i,q,Aiq);
    }
}

void Jtimes(gsl_matrix* A, int p, int q, double theta) {
    double c=cos(theta),s=sin(theta);
    for(int i=0;i<A->size2;i++) {
        double Api = gsl_matrix_get(A,p,i)*c+gsl_matrix_get(A,q,i)*s;
        double Aqi = gsl_matrix_get(A,q,i)*c-gsl_matrix_get(A,p,i)*s;
        gsl_matrix_set(A,p,i,Api);
        gsl_matrix_set(A,q,i,Aqi);
    }
}

void jacobi_diag(gsl_matrix* A, gsl_matrix* V) {
    gsl_matrix_set_identity(V);
    assert(A->size1==A->size2);
    for(int i=0;i<A->size1;i++) {for(int j=i;j<A->size2;j++) {assert(gsl_matrix_get(A,i,j)==gsl_matrix_get(A,j,i));}}
    int changed = 1;
    while(changed!=0) {
        changed = 0;
        for(int i=0;i<A->size1;i++) {
            for(int j=i+1;j<A->size1;j++) {
                double Aii = gsl_matrix_get(A,i,i);
                double Ajj = gsl_matrix_get(A,j,j);
                double phi = 0.5*atan2(2*gsl_matrix_get(A,i,j),Ajj-Aii);
                Jtimes(A,i,j,-phi);
                timesJ(A,i,j,phi);
                timesJ(V,i,j,phi);
                if(Aii!=gsl_matrix_get(A,i,i) || Ajj!=gsl_matrix_get(A,j,j)) {changed = 1;}
            }
        }
    }
}

void timesJ_o(gsl_matrix* A, int p, int q, double theta) {
    double c=cos(theta),s=sin(theta);
    for(int i=0;i<=q;i++) {
        if(p-i>=0) {
            double Aip = gsl_matrix_get(A,i,p)*c-gsl_matrix_get(A,i,q)*s;
            double Aiq = gsl_matrix_get(A,i,p)*s+gsl_matrix_get(A,i,q)*c;
            gsl_matrix_set(A,i,p,Aip);
            gsl_matrix_set(A,i,q,Aiq);
        }
        else {
            double Aiq = gsl_matrix_get(A,i,p)*s+gsl_matrix_get(A,i,q)*c;
            gsl_matrix_set(A,i,q,Aiq);
        }
    }
}

void Jtimes_o(gsl_matrix* A, int p, int q, double theta) {
    double c=cos(theta),s=sin(theta);
    for(int i=p;i<A->size2;i++) {
        if(i-q<0) {
            double Api = gsl_matrix_get(A,p,i)*c+gsl_matrix_get(A,q,i)*s;
            gsl_matrix_set(A,p,i,Api);
        }
        else {
            double Api = gsl_matrix_get(A,p,i)*c+gsl_matrix_get(A,q,i)*s;
            double Aqi = gsl_matrix_get(A,q,i)*c-gsl_matrix_get(A,p,i)*s;
            gsl_matrix_set(A,p,i,Api);
            gsl_matrix_set(A,q,i,Aqi);
        }
    }
}

void jacobi_diag_o(gsl_matrix* A, gsl_matrix* V) {
    gsl_matrix_set_identity(V);
    assert(A->size1==A->size2);
    for(int i=0;i<A->size1;i++) {for(int j=i;j<A->size2;j++) {assert(gsl_matrix_get(A,i,j)==gsl_matrix_get(A,j,i));}}
    int changed = 1;
    while(changed!=0) {
        changed = 0;
        for(int i=0;i<A->size1;i++) {
            for(int j=i+1;j<A->size1;j++) {
                double Aii = gsl_matrix_get(A,i,i);
                double Ajj = gsl_matrix_get(A,j,j);
                double phi = 0.5*atan2(2*gsl_matrix_get(A,i,j),Ajj-Aii);
                Jtimes_o(A,i,j,-phi);
                timesJ_o(A,i,j,phi);
                timesJ_o(V,i,j,phi);
                if(Aii!=gsl_matrix_get(A,i,i) || Ajj!=gsl_matrix_get(A,j,j)) {changed = 1;}
            }
        }
    }
}