#include "functions.h"

int main() {
    //Diagonalizationtest -------------------------------------------------------------
    int N = 5;
    gsl_matrix* A = gsl_matrix_alloc(N,N);
    gsl_matrix* D = gsl_matrix_alloc(N,N);
    gsl_matrix* V = gsl_matrix_alloc(N,N);
    gsl_matrix* VtAV = gsl_matrix_alloc(N,N);
    gsl_matrix* AV = gsl_matrix_alloc(N,N);
    gsl_matrix* VDVt = gsl_matrix_alloc(N,N);
    gsl_matrix* DVt = gsl_matrix_alloc(N,N);
    gsl_matrix* VtV = gsl_matrix_alloc(N,N);
    for(int i=0;i<N;i++) {
        for(int j=i;j<N;j++) {
            double Aij = RND;
            gsl_matrix_set(A,i,j,Aij);
            gsl_matrix_set(A,j,i,Aij);
        }
    }
    gsl_matrix_memcpy(D,A);
    FILE* diag_test_stream = fopen("diagonalizationtest.txt","w");
    fprintf(diag_test_stream,"Matrix A before diagonalization:\n");
    print_matrix(A,diag_test_stream);
    fprintf(diag_test_stream,"Matrix D after diagonalization:\n");
    jacobi_diag(D,V);
    print_matrix(D,diag_test_stream);
    fprintf(diag_test_stream,"Eigenvector matrix V:\n");
    print_matrix(V,diag_test_stream);
    fprintf(diag_test_stream,"The product VtAV which should be equal to D:\n");
    gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1,A,V,0,AV);
    gsl_blas_dgemm(CblasTrans,CblasNoTrans,1,V,AV,0,VtAV);
    print_matrix(VtAV,diag_test_stream);
    fprintf(diag_test_stream,"The product VDVt which should be equal to A:\n");
    gsl_blas_dgemm(CblasNoTrans,CblasTrans,1,D,V,0,DVt);
    gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1,V,DVt,0,VDVt);
    print_matrix(VDVt,diag_test_stream);
    fprintf(diag_test_stream,"The product VtV which should be equal to I:\n");
    gsl_blas_dgemm(CblasTrans,CblasNoTrans,1,V,V,0,VtV);
    print_matrix(VtV,diag_test_stream);
    fclose(diag_test_stream);

    gsl_matrix_free(A);
    gsl_matrix_free(D);
    gsl_matrix_free(V);
    gsl_matrix_free(VtAV);
    gsl_matrix_free(AV);
    gsl_matrix_free(VDVt);
    gsl_matrix_free(DVt);
    gsl_matrix_free(VtV);

    //Numerically solving a quantum particle in a box ------------------------------------------
    int n = 100;
    double s = 1.0/(n+1);
    gsl_matrix* H = gsl_matrix_alloc(n,n);
    gsl_matrix* Vp = gsl_matrix_alloc(n,n);
    for(int i=0;i<n-1;i++){
        gsl_matrix_set(H,i,i,-2);
        gsl_matrix_set(H,i,i+1,1);
        gsl_matrix_set(H,i+1,i,1);
    }
    gsl_matrix_set(H,n-1,n-1,-2);
    gsl_matrix_scale(H,-1.0/s/s);
    jacobi_diag(H,Vp);
    FILE* particle_stream = fopen("particle_solve.txt","w");
    for(int k=1;k<=n;k++) {
        double exact_value = M_PI*M_PI*k*k;
        double num_value = gsl_matrix_get(H,k-1,k-1);
        fprintf(particle_stream,"n: %5i | Theoretical value: %10g | Numerical value: %10g\n",k,exact_value,num_value);
    }
    fclose(particle_stream);
    FILE* func_data_stream = fopen("num_data.txt","w");
    for(int i=0;i<n;i++) {
        fprintf(func_data_stream,"%i %g %g %g %g\n",
        i,-gsl_matrix_get(Vp,i,0),-gsl_matrix_get(Vp,i,1),-gsl_matrix_get(Vp,i,2),-gsl_matrix_get(Vp,i,3));
    }
    fclose(func_data_stream);

    gsl_matrix_free(H);
    gsl_matrix_free(Vp);

    //Time test  -----------------------------------------------------------------
    FILE* time_test_out_stream = fopen("time_test.txt","w");
    for(int N=50;N<150;N+=2) {
        printf("%i\n",N);
        gsl_matrix* My_time_test_A = gsl_matrix_alloc(N,N);
        gsl_matrix* My_time_test_V = gsl_matrix_alloc(N,N);
        //gsl_matrix* My_time_test_A_o = gsl_matrix_alloc(N,N);
        //gsl_matrix* My_time_test_V_o = gsl_matrix_alloc(N,N);
        gsl_matrix* Gsl_time_test_A = gsl_matrix_alloc(N,N);  
        gsl_matrix* Gsl_time_test_v = gsl_matrix_alloc(N,N);
        gsl_vector* Gsl_time_test_eval = gsl_vector_alloc(N);
        for(int i=0;i<N;i++) {
            for(int j=i;j<N;j++) {
                double Aij = RND;
                gsl_matrix_set(My_time_test_A,i,j,Aij);
                gsl_matrix_set(My_time_test_A,j,i,Aij);
            }
        }
        gsl_matrix_memcpy(Gsl_time_test_A,My_time_test_A);
        //gsl_matrix_memcpy(My_time_test_A_o,My_time_test_A);

        clock_t my_time1 = clock();
        jacobi_diag(My_time_test_A,My_time_test_V);
        clock_t my_time2 = clock();
        double my_time = (double)(my_time2-my_time1);
        /* Couldnt get optimized version working
        clock_t my_time1_o = clock();
        jacobi_diag_o(My_time_test_A_o,My_time_test_V_o);
        clock_t my_time2_o = clock();
        double my_time_o = (double)(my_time2_o-my_time1_o);
        */
        clock_t gsl_time1 = clock();
        gsl_eigen_symmv_workspace* gsl_ws = gsl_eigen_symmv_alloc(N);
        gsl_eigen_symmv(Gsl_time_test_A,Gsl_time_test_eval,Gsl_time_test_v,gsl_ws);
        gsl_eigen_symmv_free(gsl_ws);
        clock_t gsl_time2 = clock();
        double gsl_time = (double)(gsl_time2-gsl_time1);

        gsl_matrix_free(My_time_test_A);
        gsl_matrix_free(My_time_test_V);
        //gsl_matrix_free(My_time_test_A_o);
        //gsl_matrix_free(My_time_test_V_o);
        gsl_matrix_free(Gsl_time_test_A);
        gsl_matrix_free(Gsl_time_test_v);
        gsl_vector_free(Gsl_time_test_eval);

        my_time /= CLOCKS_PER_SEC;
        //my_time_o /= CLOCKS_PER_SEC;
        gsl_time /= CLOCKS_PER_SEC;
        double func_val = 4.5e-7*pow(N,3);
        fprintf(time_test_out_stream,"%i %g %g %g\n",N,my_time,gsl_time,func_val);
    }
    fclose(time_test_out_stream);

return 0;
}