#include<stdio.h>
#include<string.h>
#include<math.h>
#include<gsl/gsl_vector.h>
#include<gsl/gsl_matrix.h>
#include<stdlib.h>
#include<gsl/gsl_linalg.h>
#include<gsl/gsl_blas.h>
#include<assert.h>
#include "functions.h"

//FUNCTIONS TO INTEGRATE//
double f(double x)
{
    return 1/sqrt(x);
}
double g(double x)
{
    return 4*sqrt(1-pow(x,2));
}
double h(double x)
{
    return exp(-pow(x,2));
}
//VARIABLE TRANSFORMATIONS//
double x(double t, double a, double b)
{
    return a+(b-a)*t;
}
double w(double w, double a, double b)
{
    return (b-a)*w;
}
double CC(double f(double), double x, double a, double b)
{
    return f((a+b)/2+(b-a)/2*cos(x))*sin(x)*(b-a)/2;
}
double a_inf(double f(double), double x, double b)
{
    return f(b-(1-x)/x)/pow(x,2);
}
double b_inf(double f(double), double x, double a)
{
    return f(a+(1-x)/x)/pow(x,2);
}
double both_inf(double f(double), double x)
{
    return (f((1-x)/x)+f(-(1-x)/x))/pow(x,2);
}
//RECURSIVE INTEGRATION FUNCTION//
double driver(double f(double), double org_a, double org_b, double a, double b, double acc, double eps,\
              double two, double three, double* high, double* low, double* xs, int* evals, double* error, char* type)
{
    double one, four;
    if(type=="normal") {
        one = f(x(xs[0],a,b));
        four = f(x(xs[3],a,b));
    }
    if(type=="a_inf") {
        one = a_inf(f,x(xs[0],a,b),org_b);
        four = a_inf(f,x(xs[3],a,b),org_b);
    }
    if(type=="b_inf") {
        one = b_inf(f,x(xs[0],a,b),org_a);
        four = b_inf(f,x(xs[3],a,b),org_a);
    }
    if(type=="both_inf") {
        one = both_inf(f,x(xs[0],a,b));
        four = both_inf(f,x(xs[3],a,b));
    }
    if(type=="CC") {
        one = CC(f,x(xs[0],a,b),org_a,org_b);
        four = CC(f,x(xs[3],a,b),org_a,org_b);
    }
    double Q = w(high[0],a,b)*one+w(high[1],a,b)*two+w(high[2],a,b)*three+w(high[3],a,b)*four; 
    double q = w(low[0],a,b)*one+w(low[1],a,b)*two+w(low[2],a,b)*three+w(low[3],a,b)*four;
    double err=fabs(Q-q);
    if (err < acc+eps*fabs(Q))
        {
            *error += pow(err,2);
            return Q;
        }
    else 
        {
            *evals += 2;
            return driver(f,org_a,org_b,a,(a+b)/2,acc/sqrt(2.0),eps,one,two,high,low,xs,evals,error,type)+
                    driver(f,org_a,org_b,(a+b)/2,b,acc/sqrt(2.0),eps,three,four,high,low,xs,evals,error,type);
        }
}
//INITIALIZING FUNCTION//
returns integrate(double f(double), double a, double b, double acc, double eps, char* is_CC) 
{
    assert(is_CC=="CC" || is_CC=="no_CC" && "is_CC param must be CC or no_CC");
    if(is_CC=="CC") {assert(a!=-INFINITY && b!=INFINITY && "CC not available with infinite boundaries");}
    double Higher_order[4] = {2.0/6,1.0/6,1.0/6,2.0/6};
    double Lower_order[4] = {1.0/4,1.0/4,1.0/4,1.0/4};
    double interval_points[4] = {1.0/6,2.0/6,4.0/6,5.0/6};
    double estimate = 0;
    double error = 0;
    int evals = 0;

    if(is_CC=="no_CC" && a!=-INFINITY && b!=INFINITY) {
        estimate = driver(f,a,b,a,b,acc,eps,f(x(interval_points[1],a,b)),f(x(interval_points[2],a,b)),\
                          Higher_order,Lower_order,interval_points,&evals,&error,"normal");
    }
    if(is_CC=="no_CC" && a==-INFINITY && b!=INFINITY) {
        estimate = driver(f,a,b,0,1,acc,eps,a_inf(f,x(interval_points[1],0,1),b),a_inf(f,x(interval_points[2],0,1),b),\
                          Higher_order,Lower_order,interval_points,&evals,&error,"a_inf");
    }
    if(is_CC=="no_CC" && a!=-INFINITY && b==INFINITY) {
        estimate = driver(f,a,b,0,1,acc,eps,b_inf(f,x(interval_points[1],0,1),a),b_inf(f,x(interval_points[2],0,1),a),\
                          Higher_order,Lower_order,interval_points,&evals,&error,"b_inf");
    }
    if(is_CC=="no_CC" && a==-INFINITY && b==INFINITY) {
        estimate = driver(f,a,b,0,1,acc,eps,both_inf(f,x(interval_points[1],0,1)),both_inf(f,x(interval_points[2],0,1)),\
                          Higher_order,Lower_order,interval_points,&evals,&error,"both_inf");
    }
    if(is_CC=="CC") {
        estimate = driver(f,a,b,0,M_PI,acc,eps,CC(f,x(interval_points[1],0,M_PI),a,b),CC(f,x(interval_points[2],0,M_PI),a,b),\
                             Higher_order,Lower_order,interval_points,&evals,&error,"CC");
    }
    
    returns returns = {estimate,sqrt(error)};
    printf("Integration(%s) took %i evals\n",is_CC,evals);
    return returns;
}
